// *************************************************************************
//    BuiltinUserRelRelation.C
// *************************************************************************

#include "BuiltinUserRelRelation.h"

#include "TupleIterator.h"
#include "Globals.h"
#include "InitBuiltins.h"
#include "Debug.h"

//#define   USER_REL_TEST
#ifdef    USER_REL_TEST
#    include "HashSimpleRelation.h"
#    include "Interface.h"
#endif

// *************************************************************************
Cor_BuiltinUserRelRelation::Cor_BuiltinUserRelRelation(int arity, Cor_CArg *n,
    UserRelProc user_op)
    : Cor_BuiltinRelation(arity, n, NULL, 0)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    op = user_op;
    single_success = 0;
    Cor_InsertBuiltin(*n,this);
}

// *************************************************************************
/*
    BuiltinUserRelRelation::get_next  is called only the first time.
    This computes the desired relation, and then destructively
    updates the tuple iterator, so that the `relation' field points
    to the user returned relation.  It then directly calls the 
    get_next on the tuple iterator, which should work fine since
    the built-in relation has been replaced by a fixed relation.
    Subsequent calls to get_next, also work with the modified 
    relation field.
*/

Cor_BindEnv *Cor_BuiltinUserRelRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *tuple = get_next_tuple(iterator);

    if (tuple)    // Eliminate "unused" warning
        ;    // Nothing

    if (iterator.no_match())
        return(NULL);
    return(iterator.bindenv);
}

// *************************************************************************
Cor_Tuple *Cor_BuiltinUserRelRelation::get_next_tuple(
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BindEnv       *env = iterator.bindenv;
    Cor_BindEnv       *queryenv = NULL;
    Cor_Relation      *rel;
    Cor_TupleIterator *iter;

    if (iterator.ipos++ == 0) {

        if (iterator.arg_list.arity() != arity()) {
            iterator.set_no_match(); 
            return(NULL);
        }

        // first create the query tuple from the bindenv
        Cor_ArgList *alist = new Cor_ArgList(arity());

        for (int i = 0; i < arity(); i++) {
            Cor_Term term(iterator.arg_list[i], env);
            Cor_full_dereference_term(term);
            if (!((term.expr.is_var() && (term.bindenv == env))
                        || term.isConstant())) {
                term.expr = iterator.arg_list[i];  // Dereferencing lead to a
                // different bindenv -- cannot use dereferenced term.expr
                // with iterator.bindenv, so revert to original arg_list[i]
                queryenv = iterator.bindenv;      
            }
            (*alist)[i] = term.dereference().expr;
        }
    
        Cor_Tuple query(alist);
        query.instantiate(queryenv); 
    
    
        // now call the user routine to compute the (single) answer to the query
        rel = (Cor_Relation *) (*op)(&query); 
    
        Cor_StackMark      stack_mark;
        stack_mark.pop_to();

        if (rel == NULL) {
            // no answer to query
            iterator.set_no_match();
            return(NULL);
        }
    
        iter = new Cor_TupleIterator(rel, iterator.arg_list, env);

        iterator.answer_iterator = iter;
        iter->state = iterator.state;

    } else {
        iter = iterator.answer_iterator;
    }

    Cor_Tuple *ret_tuple = iter->get_next_tuple();

    if (iter->no_match()) {
        iterator.set_no_match();
        return(NULL);
    } else {
        iterator.reset_no_match();
        return(ret_tuple);
    }
}

// *************************************************************************

#ifdef    USER_REL_TEST
// Sample addition of a user defined builtin relation.

Cor_Relation *substr_solver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
	if (query == NULL) {
	    cerr << "Null argument passed to substr_solver.\n"; 
	    return NULL; 
	} 

	int arity = query->arity(); 
    if (arity != 4) {
	    cerr << "substr() requires 4 arguments.\n";
        return(NULL);
    }

    Cor_Relation *rel = new Cor_HashSimpleRelation(arity);

    rel->check_subsum = 0;
    rel->name = Cor_CArg("__SUBSTR__");

    char *seq = (query->args())[0].symbol();
    int   seq_len = strlen(seq);

    int i1, i2;

    char         sv_char;    // To assist in cheating
    Cor_Tuple   *tuple;
	Cor_ArgList *alist;

    // o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    // Solve for all cases  (bbbf, bbff, bfbf, bfff)

    if ((query->args())[1].is_int()) {       // i1 -bound

        i1 = make_int((query->args())[1]);

        // o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
        //  bbbf

        if ((query->args())[2].is_int()) {   // i1, i2 -bound

            i2 = make_int((query->args())[2]);

	        alist = new Cor_ArgList(arity);
	        (*alist)[0] = (query->args())[0];
	        (*alist)[1] = (query->args())[1];
	        (*alist)[2] = (query->args())[2];

            if (i2 < i1) {
                (*alist)[3] = Cor_CArg("");
            } else {
                sv_char = seq[i2];
                seq[i2] = '\0';

                (*alist)[3] = Cor_CArg(&(seq[i1-1]));
                seq[i2] = sv_char;        // Restore the original string
            }

            tuple = new Cor_Tuple(alist);
            rel->insert_tuple(tuple);

        // o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
        //  bbff

        } else {                             // i1 -bound, i2 -not bound

            for (i2 = i1-1; i2 <= seq_len; i2++) {

    	        alist = new Cor_ArgList(arity);
    	        (*alist)[0] = (query->args())[0];
    	        (*alist)[1] = (query->args())[1];
    	        (*alist)[2] = i2;
    
                if (i2 < i1) {
                    (*alist)[3] = Cor_CArg("");
                } else {
                    sv_char = seq[i2];
                    seq[i2] = '\0';
                    (*alist)[3] = Cor_CArg(&(seq[i1-1]));
                    seq[i2] = sv_char;        // Restore the original string
                }
                tuple = new Cor_Tuple(alist);
                rel->insert_tuple(tuple);
            }
        }

    } else {                                 // i1 -not bound

        // o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
        //  bfbf

        if ((query->args())[2].is_int()) {   // i1 -not bound, i2 -bound

            i2 = make_int((query->args())[2]);

            for (i1 = 1;    (i1<i2+1) && (i1 <= seq_len+1); i1++) {

    	        alist = new Cor_ArgList(arity);
    	        (*alist)[0] = (query->args())[0];
    	        (*alist)[1] = i1;
    	        (*alist)[2] = (query->args())[2];
    
                if (i2 < i1) {
                    (*alist)[3] = Cor_CArg("");
                } else {
                    sv_char = seq[i2];
                    seq[i2] = '\0';
    
                    (*alist)[3] = Cor_CArg(&(seq[i1-1]));
                    seq[i2] = sv_char;        // Restore the original string
                }
    
                tuple = new Cor_Tuple(alist);
                rel->insert_tuple(tuple);
            }

        // o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
        //  bfff

        } else {                             // i1 -not bound, i2 -not bound

            for (i1 = 1; i1 <= seq_len+1; i1++) {
                for (i2 = i1-1; i2 <= seq_len; i2++) {
    
        	        alist = new Cor_ArgList(arity);
        	        (*alist)[0] = (query->args())[0];
        	        (*alist)[1] = i1;
        	        (*alist)[2] = i2;
        
                    if (i2 < i1) {
                        (*alist)[3] = Cor_CArg("");
                    } else {
                        sv_char = seq[i2];
                        seq[i2] = '\0';
        
                        (*alist)[3] = Cor_CArg(&(seq[i1-1]));
                        seq[i2] = sv_char;        // Restore the original string
                    }
        
                    tuple = new Cor_Tuple(alist);
                    rel->insert_tuple(tuple);
                }
            }
        }
    }

    return(rel);
}
#endif    USER_REL_TEST

// *************************************************************************
