// *************************************************************************
//    GoalNode.C
// *************************************************************************

#include "GoalNode.h"

#include "RelationMisc.h"
#include "Util.h"
#include "StorageRelation.h"
#include "Execute.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************
Cor_GoalNode::Cor_GoalNode(Cor_StorageRelation *rel, int neg) : Cor_Tuple()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    set_defaults(rel, neg);
}

// *************************************************************************
Cor_GoalNode::Cor_GoalNode(int arity, Cor_StorageRelation *rel,
        int neg): Cor_Tuple(arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    set_defaults(rel, neg);
}

// *************************************************************************
Cor_GoalNode::Cor_GoalNode(Cor_ArgList *arglist, Cor_BindEnv *env,
        Cor_StorageRelation *rel,
    int neg)
    : Cor_Tuple(arglist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    instantiate(env);
    set_defaults(rel, neg);
}

// *************************************************************************
Cor_GoalNode::Cor_GoalNode(Cor_ArgList&a, Cor_StorageRelation *rel,
        int neg) : Cor_Tuple(a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    set_defaults(rel, neg);
}

// *************************************************************************
Cor_GoalNode::Cor_GoalNode(Cor_ArgList *arglist, Cor_StorageRelation *rel,
        int neg) : Cor_Tuple(arglist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    set_defaults(rel, neg);
}

// *************************************************************************
Cor_GoalNode::~Cor_GoalNode()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    //cerr << "In delete for GoalNode, dropping context: " << c_node << endl;
    //cerr << "In ~Cor_GoalNode this is " << this << endl;
}

// *************************************************************************
void Cor_GoalNode::set_defaults(Cor_StorageRelation *rel, int neg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    mark = 0;
    parent_goal = NULL;
    c_node = NULL;
    owner_rel = rel;

    negated = neg;

    // now set the fields related to makeset() of union-find
    parent_in_set = this;
    rank_in_set = 0;
}

// *************************************************************************
Cor_GoalNode *Cor_GoalNode::find_canonical_elem ()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // this is the find() operation in tarjan's union-find operation
    // it automatically does path compression, while doing a single
    // pass through the parent_in_set chain

    if (this != parent_in_set) {
        parent_in_set = parent_in_set->find_canonical_elem();
    }

    return(parent_in_set);
}

// *************************************************************************
Cor_ContextNode *Cor_GoalNode::find_context_node()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // GoalNode maintains a c_node field, but for reasons of 
    // asymptotics, only the c_node field in the canonical GoalNode
    // of the set is the CORRECT one.

    Cor_GoalNode *can_g_node = find_canonical_elem();

    return(can_g_node->c_node);
}

// *************************************************************************
//  NOTE: unmarked goal nodes in marked context nodes better be 
//  guaranteed to be leaf nodes because this assumption is used elsewhere
// *************************************************************************
Cor_GoalNode *Cor_GoalNode::link_goals(Cor_GoalNode *g)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // this is the link() operation of tarjan's union-find
    if (rank_in_set < g->rank_in_set) {
        parent_in_set = g;
        g->c_node = c_node; // since g is the root of merged set
        return(g);
    }
    if ((rank_in_set = g->rank_in_set)) {
        // going to make "this" as g's parent 
        rank_in_set++;
    }
    g->parent_in_set = this;
    g->c_node = NULL;                // for safety only
    return(this);
}

// *************************************************************************
int Cor_GoalNode::is_magic_goal()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(owner_rel->r_kind == COR_R_MAGIC);
}

// *************************************************************************
Cor_StorageRelation *Cor_GoalNode::find_done_relation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT((is_magic_goal()));
    if (!(owner_rel->mp_done_reln)) {
        char *rel_str = owner_rel->name.symbol();
        int rel_str_len = strlen(rel_str);

        char *done_str = Cor_make_done_prefix();
        int done_str_len = strlen(done_str);

        char *rel_done_str = new char [rel_str_len+done_str_len+1];

        strcpy(rel_done_str, done_str);
        strcpy(rel_done_str+done_str_len, rel_str);
        rel_done_str[done_str_len+rel_str_len] = '\0';
        
        Cor_CArg rel_done_name = Cor_CArg(rel_done_str);
        owner_rel->mp_done_reln = (Cor_StorageRelation *)
            Cor_find_local_relation(rel_done_name, owner_rel->arity());
            // this only searches on the current frame

        if (!owner_rel->mp_done_reln) {
            // not on frame
            owner_rel->mp_done_reln = (Cor_StorageRelation *)
                Cor_make_local_relation(rel_done_name, owner_rel->arity(), 0);
                // NOTE: where can this relation be deleted???
        }

        delete [] rel_done_str;        // free up some space here
    }
    return(owner_rel->mp_done_reln);
}

// *************************************************************************
void Cor_GoalNode::print_all(ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *file << "\t\tPrinting GoalNode information.\n";
    *file << "\t\tMark = " << mark << ";    ";
    *file << owner_rel->name.symbol() << " (";
    Cor_Tuple::printon(file);
    *file << ")\n";
}

// *************************************************************************
